@@ -29,8 +29,7 @@ import re
 import types
 import warnings
 import zlib
-
-from itertools import imap
+import string
 
 from twisted.internet import defer
 from twisted.internet import protocol
@@ -82,27 +81,50 @@ def list_or_args(command, keys, args):
         keys.extend(args)
     return keys
 
+# Possible first characters in a string containing an integer or a float.
+_NUM_FIRST_CHARS = frozenset(string.digits + "+-.")
+
+
+class MultiBulkStorage(object):
+    def __init__(self, parent=None):
+        self.items = None
+        self.pending = None
+        self.parent = parent
+
+    def set_pending(self, pending):
+        if self.pending is None:
+            if pending < 0:
+                self.items = None
+                self.pending = 0
+            else:
+                self.items = []
+                self.pending = pending
+            return self
+        else:
+            m = MultiBulkStorage(self)
+            m.set_pending(pending)
+            return m
+
+    def append(self, item):
+        self.pending -= 1
+        self.items.append(item)
+
 
 class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
     """
     Redis client protocol.
     """
 
-    ERROR = "-"
-    STATUS = "+"
-    INTEGER = ":"
-    BULK = "$"
-    MULTI_BULK = "*"
-
     def __init__(self, charset="utf-8", errors="strict"):
         self.charset = charset
         self.errors = errors
 
         self.bulk_length = 0
-        self.bulk_buffer = ""
-        self.multi_bulk_length = 0
-        self.multi_bulk_reply = []
-        self.replyQueueLength = 0
+        self.bulk_buffer = []
+
+        self.post_proc = []
+        self.multi_bulk = MultiBulkStorage()
+
         self.replyQueue = defer.DeferredQueue()
 
         self.transactions = 0
@@ -133,7 +155,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         self.connected = 0
         self.factory.delConnection(self)
         basic.LineReceiver.connectionLost(self, why)
-        while self.replyQueueLength < 0:
+        while self.replyQueue.pending:
             self.replyReceived(ConnectionError("Lost connection"))
 
     def lineReceived(self, line):
@@ -145,46 +167,67 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
           "$" bulk data
           "*" multi-bulk data
         """
-        if not line:
+        if line:
+            self.resetTimeout()
+            token, data = line[0], line[1:]
+        else:
             return
 
-        self.resetTimeout()
-        token = line[0]  # first byte indicates reply type
-        data = line[1:]
-        if token == self.ERROR:
-            self.errorReceived(data)
-        elif token == self.STATUS:
-            self.statusReceived(data)
-        elif token == self.INTEGER:
-            self.integerReceived(data)
-        elif token == self.BULK:
+        if token == "$":  # bulk data
             try:
                 self.bulk_length = long(data)
             except ValueError:
-                self.replyReceived(InvalidResponse(
-                 "Cannot convert data '%s' to integer" % data))
-                return
-            if self.bulk_length == -1:
-                self.bulkDataReceived(None)
-                return
+                self.replyReceived(InvalidResponse("Cannot convert data "
+                                                   "'%s' to integer" % data))
             else:
-                self.bulk_length += 2  # \r\n
-                self.setRawMode()
-        elif token == self.MULTI_BULK:
+                if self.bulk_length == -1:
+                    self.bulk_length = 0
+                    self.bulkDataReceived(None)
+                else:
+                    self.bulk_length += 2  # 2 == \r\n
+                    self.setRawMode()
+
+        elif token == "*":  # multi-bulk data
             try:
-                self.multi_bulk_length = long(data)
+                n = long(data)
             except (TypeError, ValueError):
-                self.replyReceived(InvalidResponse(
-                 "Cannot convert multi-response header '%s' to integer" % \
-                 data))
-                self.multi_bulk_length = 0
-                return
-            if self.multi_bulk_length == -1:
-                self.multi_bulk_reply = None
-                self.multiBulkDataReceived()
-                return
-            elif self.multi_bulk_length == 0:
-                self.multiBulkDataReceived()
+                self.multi_bulk = MultiBulkStorage()
+                self.replyReceived(InvalidResponse("Cannot convert "
+                                                   "multi-response header "
+                                                   "'%s' to integer" % data))
+            else:
+                self.multi_bulk = self.multi_bulk.set_pending(n)
+                if n in (0, -1):
+                    self.multiBulkDataReceived()
+
+        elif token == "+":  # single line status
+            if data == "QUEUED":
+                self.transactions += 1
+                self.replyReceived(data)
+            else:
+                if self.multi_bulk.pending:
+                    self.handleMultiBulkElement(data)
+                else:
+                    self.replyReceived(data)
+
+        elif token == "-":  # error
+            reply = ResponseError(data[4:] if data[:4] == "ERR" else data)
+            if self.multi_bulk.pending:
+                self.handleMultiBulkElement(reply)
+            else:
+                self.replyReceived(reply)
+
+        elif token == ":":  # integer
+            try:
+                reply = int(data)
+            except ValueError:
+                reply = InvalidResponse(
+                        "Cannot convert data '%s' to integer" % data)
+
+            if self.multi_bulk.pending:
+                self.handleMultiBulkElement(reply)
+            else:
+                self.replyReceived(reply)
 
     def rawDataReceived(self, data):
         """
@@ -196,151 +239,120 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         else:
             rest = ""
 
-        self.bulk_buffer += data
+        self.bulk_buffer.append(data)
         if self.bulk_length == 0:
-            bulk_buffer = self.bulk_buffer[:-2]
-            self.bulk_buffer = ""
+            bulk_buffer = "".join(self.bulk_buffer)[:-2]
+            self.bulk_buffer = []
             self.bulkDataReceived(bulk_buffer)
-            while self.multi_bulk_length > 0 and rest:
-                if rest[0] == self.BULK:
-                    idx = rest.find(self.delimiter)
-                    if idx == -1:
-                        break
-                    data_len = int(rest[1:idx], 10)
-                    if data_len == -1:
-                        rest = rest[5:]
-                        self.bulkDataReceived(None)
-                        continue
-                    elif len(rest) >= (idx + 5 + data_len):
-                        data_start = idx + 2
-                        data_end = data_start + data_len
-                        data = rest[data_start: data_end]
-                        rest = rest[data_end + 2:]
-                        self.bulkDataReceived(data)
-                        continue
-                break
-            self.setLineMode(extra=rest)
-
-    def errorReceived(self, data):
-        """
-        Error from server.
-        """
-        reply = ResponseError(data[4:] if data[:4] == "ERR " else data)
-
-        if self.multi_bulk_length:
-            self.handleMultiBulkElement(reply)
-        else:
-            self.replyReceived(reply)
-
-    def statusReceived(self, data):
-        """
-        Single line status should always be a string.
-        """
-        #if data == "none":
-        #    reply = None # should this happen here in the client?
-        #else:
-        #    reply = data
-
-        reply = data
-        if reply == "QUEUED":
-            self.transactions += 1
-            self.replyReceived(reply)
-            return
-
-        if self.multi_bulk_length:
-            self.handleMultiBulkElement(reply)
-        else:
-            self.replyReceived(reply)
-
-    def integerReceived(self, data):
-        """
-        For handling integer replies.
-        """
-        try:
-            reply = int(data)
-        except ValueError:
-            reply = InvalidResponse(
-            "Cannot convert data '%s' to integer" % data)
-
-        if self.multi_bulk_length:
-            self.handleMultiBulkElement(reply)
-        else:
-            self.replyReceived(reply)
+            reactor.callLater(0, self.setLineMode, extra=rest)
 
     def bulkDataReceived(self, data):
         """
         Receipt of a bulk data element.
         """
-        if data is None:
-            element = data
-        else:
-            try:
-                element = int(data) if data.find('.') == -1 else float(data)
-            except (ValueError):
+        el = None
+        if data is not None:
+            if data and data[0] in _NUM_FIRST_CHARS:  # Most likely a number
                 try:
-                    element = data.decode(self.charset)
+                    el = int(data) if data.find('.') == -1 else float(data)
+                except ValueError:
+                    pass
+
+            if el is None:
+                try:
+                    el = data.decode(self.charset)
                 except UnicodeDecodeError:
-                    element = data
-        if self.multi_bulk_length > 0:
-            self.handleMultiBulkElement(element)
+                    el = data
+
+        if self.multi_bulk.pending or self.multi_bulk.items:
+            self.handleMultiBulkElement(el)
         else:
-            self.replyReceived(element)
+            self.replyReceived(el)
 
     def handleMultiBulkElement(self, element):
-        if self.inTransaction:
-            self.transactions -= 1
-        self.multi_bulk_reply.append(element)
-        self.multi_bulk_length = self.multi_bulk_length - 1
+        self.multi_bulk.append(element)
 
-        if self.multi_bulk_length == 0:
+        if not self.multi_bulk.pending:
             self.multiBulkDataReceived()
 
     def multiBulkDataReceived(self):
         """
         Receipt of list or set of bulk data elements.
         """
-        if self.inTransaction and self.transactions == 0:
-            self.inTransaction = False
-        reply = self.multi_bulk_reply
-        self.multi_bulk_reply = []
-        self.multi_bulk_length = 0
-        self.replyReceived(reply)
+        while self.multi_bulk.parent and not self.multi_bulk.pending:
+            p = self.multi_bulk.parent
+            p.append(self.multi_bulk.items)
+            self.multi_bulk = p
+
+        if not self.multi_bulk.pending:
+            reply = self.multi_bulk.items
+            self.multi_bulk = MultiBulkStorage()
+
+            if self.inTransaction and reply is not None:
+                self.transactions -= len(reply)
+                if not self.transactions:
+                    self.inTransaction = False
+
+                    tmp = []
+                    for f, v in zip(self.post_proc[1:], reply):
+                        if callable(f):
+                            tmp.append(f(v))
+                        else:
+                            tmp.append(v)
+                        reply = tmp
+                self.post_proc = []
+
+            self.replyReceived(reply)
 
     def replyReceived(self, reply):
         """
         Complete reply received and ready to be pushed to the requesting
         function.
         """
-        self.replyQueueLength += 1
         self.replyQueue.put(reply)
 
-    def get_response(self):
-        """return deferred which will fire with response from server.
-        """
-        self.replyQueueLength -= 1
-        return self.replyQueue.get()
+    @staticmethod
+    def handle_reply(r):
+        if isinstance(r, Exception):
+            raise r
+        return r
 
-    def encode(self, s):
-        if isinstance(s, str):
-            return s
-        if isinstance(s, unicode):
-            try:
-                return s.encode(self.charset, self.errors)
-            except UnicodeEncodeError, e:
-                raise InvalidData(
-                "Error encoding unicode value '%s': %s" % (repr(s), e))
-        if isinstance(s, float):
-            return format(s, "f")
-        return str(s)
-
-    def execute_command(self, *args):
+    def execute_command(self, *args, **kwargs):
         if self.connected == 0:
             raise ConnectionError("Not connected")
         else:
-            cmds = ["$%s\r\n%s\r\n" % (len(enc_value), enc_value)
-                for enc_value in imap(self.encode, args)]
+            cmds = []
+            cmd_template = "$%s\r\n%s\r\n"
+            for s in args:
+                if isinstance(s, str):
+                    cmd = s
+                elif isinstance(s, unicode):
+                    try:
+                        cmd = s.encode(self.charset, self.errors)
+                    except UnicodeEncodeError, e:
+                        raise InvalidData(
+                            "Error encoding unicode value '%s': %s" %
+                            (repr(s), e))
+                elif isinstance(s, float):
+                    try:
+                        cmd = format(s, "f")
+                    except NameError:
+                        cmd = "%0.6f" % s
+                else:
+                    cmd = str(s)
+                cmds.append(cmd_template % (len(cmd), cmd))
             self.transport.write("*%s\r\n%s" % (len(cmds), "".join(cmds)))
-            return self.get_response()
+            r = self.replyQueue.get().addCallback(self.handle_reply)
+
+            if self.inTransaction:
+                self.post_proc.append(kwargs.get("post_proc"))
+            else:
+                if "post_proc" in kwargs:
+                    f = kwargs["post_proc"]
+                    if callable(f):
+                        r.addCallback(f)
+
+            return r
 
     ##
     # REDIS COMMANDS
@@ -671,6 +683,13 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         keys.append(timeout)
         return self.execute_command("BRPOP", *keys)
 
+    def brpoplpush(self, source, destination):
+        """
+        Pop a value from a list, push it to another list and return
+        it; or block until one is available.
+        """
+        return self.execute_command("BRPOPLPUSH", source, destination)
+
     def rpoplpush(self, srckey, dstkey):
         """
         Return and remove (atomically) the last element of the source
@@ -710,7 +729,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         Move the specified member from one Set to another atomically
         """
         return self.execute_command(
-                "SMOVE", srckey, dstkey, member).addCallback(bool)
+            "SMOVE", srckey, dstkey, member).addCallback(bool)
 
     def scard(self, key):
         """
@@ -730,7 +749,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         keys = list_or_args("sinter", keys, args)
         return self.execute_command("SINTER", *keys).addCallback(
-                self._make_set)
+            self._make_set)
 
     def sinterstore(self, dstkey, keys, *args):
         """
@@ -746,7 +765,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         keys = list_or_args("sunion", keys, args)
         return self.execute_command("SUNION", *keys).addCallback(
-                self._make_set)
+            self._make_set)
 
     def sunionstore(self, dstkey, keys, *args):
         """
@@ -763,7 +782,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         keys = list_or_args("sdiff", keys, args)
         return self.execute_command("SDIFF", *keys).addCallback(
-                self._make_set)
+            self._make_set)
 
     def sdiffstore(self, dstkey, keys, *args):
         """
@@ -778,7 +797,7 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         Return all the members of the Set value at key
         """
         return self.execute_command("SMEMBERS", key).addCallback(
-                self._make_set)
+            self._make_set)
 
     def srandmember(self, key):
         """
@@ -872,9 +891,8 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         return self._zrange(key, start, end, withscores, True)
 
-    def _zrangebyscore(self, key, min, max, withscores, offset,
-            count, reverse):
-        if reverse:
+    def _zrangebyscore(self, key, min, max, withscores, offset, count, rev):
+        if rev:
             cmd = "ZREVRANGEBYSCORE"
         else:
             cmd = "ZRANGEBYSCORE"
@@ -893,22 +911,22 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         return r
 
     def zrangebyscore(self, key, min='-inf', max='+inf', withscores=False,
-            offset=None, count=None):
+                      offset=None, count=None):
         """
         Return all the elements with score >= min and score <= max
         (a range query) from the sorted set
         """
         return self._zrangebyscore(key, min, max, withscores, offset,
-                count, False)
+                                   count, False)
 
     def zrevrangebyscore(self, key, max='+inf', min='-inf', withscores=False,
-            offset=None, count=None):
+                         offset=None, count=None):
         """
         ZRANGEBYSCORE in reverse order
         """
         # ZREVRANGEBYSCORE takes max before min
         return self._zrangebyscore(key, max, min, withscores, offset,
-                count, True)
+                                   count, True)
 
     def zcount(self, key, min='-inf', max='+inf'):
         """
@@ -1045,11 +1063,15 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         return self.execute_command("HEXISTS", key, field)
 
-    def hdel(self, key, field):
+    def hdel(self, key, fields):
         """
-        Remove the specified field from a hash
+        Remove the specified field or fields from a hash
         """
-        return self.execute_command("HDEL", key, field)
+        if isinstance(fields, (str, unicode)):
+            fields = [fields]
+        else:
+            fields = list(fields)
+        return self.execute_command("HDEL", key, *fields)
 
     def hlen(self, key):
         """
@@ -1069,19 +1091,18 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
         """
         return self.execute_command("HVALS", key)
 
-    @defer.inlineCallbacks
     def hgetall(self, key):
         """
         Return all the fields and associated values in a hash.
         """
-        d = yield self.execute_command("HGETALL", key)
-        defer.returnValue(dict(zip(d[::2], d[1::2])))
+        f = lambda d: dict(zip(d[::2], d[1::2]))
+        return self.execute_command("HGETALL", key, post_proc=f)
 
     # Sorting
     def sort(self, key, start=None, end=None, by=None, get=None,
-            desc=None, alpha=False, store=None):
+             desc=None, alpha=False, store=None):
         if (start is not None and end is not None) or \
-            (end is not None and start is None):
+           (end is not None and start is None):
             raise RedisError("``start`` and ``end`` must both be specified")
 
         pieces = [key]
@@ -1111,22 +1132,22 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
     # the transaction. At the end, either exec() or discard()
     # must be executed.
     def multi(self, keys=None):
+        self.inTransaction = True
         if keys:
             if isinstance(keys, (str, unicode)):
                 keys = [keys]
             d = defer.Deferred()
             self.execute_command("WATCH", *keys).addCallback(
-                    self._watch_added, d)
+                self._watch_added, d)
         else:
             d = self.execute_command("MULTI").addCallback(self._multi_started)
-        self.inTransaction = True
         return d
 
     def _watch_added(self, response, d):
         if response != 'OK':
             d.errback(RedisError('Invalid WATCH response: %s' % response))
         self.execute_command("MULTI").addCallback(
-                self._multi_started).chainDeferred(d)
+            self._multi_started).chainDeferred(d)
 
     def _multi_started(self, response):
         if response != 'OK':
@@ -1135,8 +1156,8 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
 
     def _commit_check(self, response):
         if response is None:
-            self.inTransaction = False
             self.transactions = 0
+            self.inTransaction = False
             raise WatchError("Transaction failed")
         else:
             return response
@@ -1144,12 +1165,12 @@ class RedisProtocol(basic.LineReceiver, policies.TimeoutMixin):
     def commit(self):
         if self.inTransaction is False:
             raise RedisError("Not in transaction")
-        return self.execute_command("EXEC").addCallback(
-                self._commit_check)
+        return self.execute_command("EXEC").addCallback(self._commit_check)
 
     def discard(self):
         if self.inTransaction is False:
             raise RedisError("Not in transaction")
+        self.post_proc = []
         self.transactions = 0
         self.inTransaction = False
         return self.execute_command("DISCARD")
@@ -1437,7 +1458,7 @@ class ShardedConnectionHandler(object):
             assert isinstance(key, (str, unicode))
         except:
             raise ValueError(
-                  "Method '%s' requires a key as the first argument" % method)
+                "Method '%s' requires a key as the first argument" % method)
 
         m = _findhash.match(key)
         if m is not None and len(m.groups()) >= 1:
@@ -1486,7 +1507,7 @@ class ShardedConnectionHandler(object):
             except:
                 pass
             else:
-                nodes.append("%s:%s/%d" % \
+                nodes.append("%s:%s/%d" %
                              (cli.host, cli.port, conn._factory.size))
         return "<Redis Sharded Connection: %s>" % ", ".join(nodes)
 
@@ -1500,7 +1521,7 @@ class ShardedUnixConnectionHandler(ShardedConnectionHandler):
             except:
                 pass
             else:
-                nodes.append("%s/%d" % \
+                nodes.append("%s/%d" %
                              (cli.name, conn._factory.size))
         return "<Redis Sharded Connection: %s>" % ", ".join(nodes)
 
@@ -1512,11 +1533,11 @@ class RedisFactory(protocol.ReconnectingClientFactory):
     def __init__(self, dbid, poolsize, isLazy=False,
                  handler=ConnectionHandler):
         if not isinstance(poolsize, int):
-            raise ValueError("Redis poolsize must be an integer, not %s" % \
+            raise ValueError("Redis poolsize must be an integer, not %s" %
                              repr(poolsize))
 
         if not isinstance(dbid, (int, types.NoneType)):
-            raise ValueError("Redis dbid must be an integer, not %s" % \
+            raise ValueError("Redis dbid must be an integer, not %s" %
                              repr(dbid))
 
         self.dbid = dbid
@@ -1713,4 +1734,4 @@ __all__ = [
 ]
 
 __author__ = "Alexandre Fiori"
-__version__ = version = "0.5"
+__version__ = version = "0.6"
